@@ -134,6 +134,7 @@ enum
 	PROP_HIGHLIGHT_SYNTAX,
 	PROP_IMPLICIT_TRAILING_NEWLINE,
 	PROP_LANGUAGE,
+	PROP_LOADING,
 	PROP_STYLE_SCHEME,
 	N_PROPERTIES
 };
@@ -165,6 +166,8 @@ typedef struct
 
 	guint cursor_moved_block_count;
 
+	int loading_count;
+
 	guint has_draw_spaces_tag : 1;
 	guint highlight_syntax : 1;
 	guint highlight_brackets : 1;
@@ -333,6 +336,25 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 				     G_PARAM_READWRITE |
 				     G_PARAM_STATIC_STRINGS);
 
+	/**
+	 * GtkSourceBuffer:loading:
+	 *
+	 * The "loading" property denotes that a `GtkSourceFileLoader` is
+	 * currently loading the buffer.
+	 *
+	 * Applications may want to use this setting to avoid doing work
+	 * while the buffer is loading such as spellchecking.
+	 *
+	 * Since: 5.10
+	 */
+	buffer_properties[PROP_LOADING] =
+		g_param_spec_boolean ("loading",
+		                      "Loading",
+		                      "If a GtkSourceFileLoader is loading the buffer",
+		                      FALSE,
+		                      G_PARAM_READABLE |
+		                      G_PARAM_STATIC_STRINGS);
+
 	/**
 	 * GtkSourceBuffer:style-scheme:
 	 *
@@ -591,6 +613,10 @@ gtk_source_buffer_get_property (GObject    *object,
 			g_value_set_object (value, priv->language);
 			break;
 
+		case PROP_LOADING:
+			g_value_set_boolean (value, gtk_source_buffer_get_loading (buffer));
+			break;
+
 		case PROP_STYLE_SCHEME:
 			g_value_set_object (value, priv->style_scheme);
 			break;
@@ -627,7 +653,7 @@ gtk_source_buffer_new (GtkTextTagTable *table)
  * gtk_source_buffer_new_with_language:
  * @language: a #GtkSourceLanguage.
  *
- * Creates a new source buffer using the highlighting patterns in `language`.  
+ * Creates a new source buffer using the highlighting patterns in `language`.
  *
  * This is equivalent to creating a new source buffer with
  * a new tag table and then calling [method@Buffer.set_language].
@@ -1405,7 +1431,7 @@ gtk_source_buffer_get_highlight_matching_brackets (GtkSourceBuffer *buffer)
  * @buffer: a #GtkSourceBuffer.
  * @highlight: %TRUE if you want matching brackets highlighted.
  *
- * Controls the bracket match highlighting function in the buffer. 
+ * Controls the bracket match highlighting function in the buffer.
  *
  * If activated, when you position your cursor over a bracket character
  * (a parenthesis, a square bracket, etc.) the matching opening or
@@ -1542,8 +1568,8 @@ gtk_source_buffer_set_language (GtkSourceBuffer   *buffer,
  * @buffer: a #GtkSourceBuffer.
  *
  * Returns the [class@Language] associated with the buffer,
- * see [method@Buffer.set_language]. 
- * 
+ * see [method@Buffer.set_language].
+ *
  * The returned object should not be unreferenced by the user.
  *
  * Returns: (nullable) (transfer none): the [class@Language] associated
@@ -1794,10 +1820,10 @@ gtk_source_buffer_real_mark_deleted (GtkTextBuffer *buffer,
  * @category: a string defining the mark category.
  * @where: location to place the mark.
  *
- * Creates a source mark in the `buffer` of category `category`.  
- * 
- * A source mark is a [class@Gtk.TextMark] but organized into categories. 
- * Depending on the category a pixbuf can be specified that will be displayed 
+ * Creates a source mark in the `buffer` of category `category`.
+ *
+ * A source mark is a [class@Gtk.TextMark] but organized into categories.
+ * Depending on the category a pixbuf can be specified that will be displayed
  * along the line of the mark.
  *
  * Like a [class@Gtk.TextMark], a [class@Mark] can be anonymous if the
@@ -1934,7 +1960,7 @@ gtk_source_buffer_forward_iter_to_source_mark (GtkSourceBuffer *buffer,
  * @category: (nullable): category to search for, or %NULL
  *
  * Moves `iter` to the position of the previous [class@Mark] of the given
- * category. 
+ * category.
  *
  * Returns %TRUE if `iter` was moved. If `category` is NULL, the
  * previous source mark can be of any category.
@@ -1967,7 +1993,7 @@ gtk_source_buffer_backward_iter_to_source_mark (GtkSourceBuffer *buffer,
  * @iter: an iterator.
  * @category: (nullable): category to search for, or %NULL
  *
- * Returns the list of marks of the given category at @iter. 
+ * Returns the list of marks of the given category at @iter.
  *
  * If @category is %NULL it returns all marks at @iter.
  *
@@ -2697,6 +2723,54 @@ compare_line_reversed (gconstpointer aptr,
 	return g_strcmp0 (b->key, a->key);
 }
 
+typedef char *(*SortKeyFunc) (const char *str,
+                              gssize      column);
+
+static char *
+sort_collate_key (const char *str,
+                  gssize      column)
+{
+	gchar *casefolded, *collated;
+
+	while (column > 0 && *str)
+	{
+		str = g_utf8_next_char (str);
+		column--;
+	}
+
+	casefolded = g_utf8_casefold (str, -1);
+	collated = g_utf8_collate_key (casefolded, -1);
+	g_free (casefolded);
+
+	return collated;
+}
+
+static char *
+sort_filename_key (const char *str,
+                   gssize      column)
+{
+	while (column > 0 && *str)
+	{
+		str = g_utf8_next_char (str);
+		column--;
+	}
+
+	return g_utf8_collate_key_for_filename (str, -1);
+}
+
+static char *
+sort_raw_key (const char *str,
+              gssize      column)
+{
+	while (column > 0 && *str)
+	{
+		str = g_utf8_next_char (str);
+		column--;
+	}
+
+	return g_strdup (str);
+}
+
 /**
  * gtk_source_buffer_sort_lines:
  * @buffer: a #GtkSourceBuffer.
@@ -2714,13 +2788,15 @@ gtk_source_buffer_sort_lines (GtkSourceBuffer    *buffer,
                               GtkSourceSortFlags  flags,
                               gint                column)
 {
+	SortKeyFunc key_func = NULL;
 	GtkTextBuffer *text_buffer;
+	GHashTable *dedup = NULL;
 	gint start_line;
 	gint end_line;
 	gint num_lines;
 	SortLine *lines;
-	gchar *last_line = NULL;
 	gint i;
+	gint j;
 
 	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
 	g_return_if_fail (start != NULL);
@@ -2756,51 +2832,48 @@ gtk_source_buffer_sort_lines (GtkSourceBuffer    *buffer,
 		return;
 	}
 
+	dedup = g_hash_table_new (g_str_hash, g_str_equal);
+
+	if ((flags & GTK_SOURCE_SORT_FLAGS_CASE_SENSITIVE) != 0)
+	{
+		key_func = sort_raw_key;
+	}
+	else if ((flags & GTK_SOURCE_SORT_FLAGS_FILENAME) != 0)
+	{
+		key_func = sort_filename_key;
+	}
+	else
+	{
+		key_func = sort_collate_key;
+	}
+
 	num_lines = end_line - start_line + 1;
 	lines = g_new0 (SortLine, num_lines);
 
-	for (i = 0; i < num_lines; i++)
+	for (i = 0, j = 0; i < num_lines; i++)
 	{
-		gchar *line;
-		gboolean free_line = FALSE;
-		glong length;
+		lines[j].line = get_line_slice (text_buffer, start_line + i);
+		lines[j].key = key_func (lines[j].line, column);
 
-		lines[i].line = get_line_slice (text_buffer, start_line + i);
+		if ((flags & GTK_SOURCE_SORT_FLAGS_REMOVE_DUPLICATES) != 0)
+		{
+			if (g_hash_table_contains (dedup, lines[j].line))
+			{
+				g_free (lines[j].line);
+				g_free (lines[j].key);
+				continue;
+			}
 
-		if ((flags & GTK_SOURCE_SORT_FLAGS_CASE_SENSITIVE) != 0)
-		{
-			line = lines[i].line;
-		}
-		else
-		{
-			line = g_utf8_casefold (lines[i].line, -1);
-			free_line = TRUE;
+			g_hash_table_add (dedup, lines[j].line);
 		}
 
-		length = g_utf8_strlen (line, -1);
-
-		if (length < column)
-		{
-			lines[i].key = NULL;
-		}
-		else if (column > 0)
-		{
-			gchar *substring;
-
-			substring = g_utf8_offset_to_pointer (line, column);
-			lines[i].key = g_utf8_collate_key (substring, -1);
-		}
-		else
-		{
-			lines[i].key = g_utf8_collate_key (line, -1);
-		}
-
-		if (free_line)
-		{
-			g_free (line);
-		}
+		j++;
 	}
 
+	num_lines = j;
+
+	g_clear_pointer (&dedup, g_hash_table_unref);
+
 	if ((flags & GTK_SOURCE_SORT_FLAGS_REVERSE_ORDER) != 0)
 	{
 		qsort (lines, num_lines, sizeof (SortLine), compare_line_reversed);
@@ -2817,16 +2890,8 @@ gtk_source_buffer_sort_lines (GtkSourceBuffer    *buffer,
 
 	for (i = 0; i < num_lines; i++)
 	{
-		if ((flags & GTK_SOURCE_SORT_FLAGS_REMOVE_DUPLICATES) != 0 &&
-		    g_strcmp0 (last_line, lines[i].line) == 0)
-		{
-			continue;
-		}
-
 		gtk_text_buffer_insert (text_buffer, start, lines[i].line, -1);
 		gtk_text_buffer_insert (text_buffer, start, "\n", -1);
-
-		last_line = lines[i].line;
 	}
 
 	gtk_text_buffer_end_user_action (text_buffer);
@@ -3106,3 +3171,46 @@ _gtk_source_buffer_unblock_cursor_moved (GtkSourceBuffer *buffer)
 		cursor_moved (buffer);
 	}
 }
+
+void
+_gtk_source_buffer_begin_loading (GtkSourceBuffer *buffer)
+{
+	GtkSourceBufferPrivate *priv = gtk_source_buffer_get_instance_private (buffer);
+
+	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
+	g_return_if_fail (priv->loading_count >= 0);
+
+	priv->loading_count++;
+
+	if (priv->loading_count == 1)
+	{
+		g_object_notify_by_pspec (G_OBJECT (buffer), buffer_properties[PROP_LOADING]);
+	}
+}
+
+void
+_gtk_source_buffer_end_loading (GtkSourceBuffer *buffer)
+{
+	GtkSourceBufferPrivate *priv = gtk_source_buffer_get_instance_private (buffer);
+
+	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
+	g_return_if_fail (priv->loading_count > 0);
+
+	priv->loading_count--;
+
+	if (priv->loading_count == 0)
+	{
+		g_object_notify_by_pspec (G_OBJECT (buffer), buffer_properties[PROP_LOADING]);
+	}
+}
+
+gboolean
+gtk_source_buffer_get_loading (GtkSourceBuffer *buffer)
+{
+	GtkSourceBufferPrivate *priv = gtk_source_buffer_get_instance_private (buffer);
+
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (priv->loading_count >= 0, FALSE);
+
+	return priv->loading_count > 0;
+}
